#include <WiFi.h>
#include <SPIFFS.h>
#include <ESPAsyncWebServer.h>
#include <AsyncJson.h>
extern "C" {
#include "libb64/cencode.h"
}

hw_timer_t *timer = NULL;
const char *ssid = "Tek-AP";
const char *password =  "12345678";
AsyncWebServer server(80);
AsyncWebSocket ws("/ws");
AsyncWebSocketClient *  mainClient = NULL;
String inputStrSerial = "";
boolean serialComplete = false;
boolean printStart = false;
boolean printStop = false;

void IRAM_ATTR onTimer(void) {
  stopTimer();
  if (!Serial.available()) {   
    serialComplete = true;
  }
}
void startTimer() {
  stopTimer();
  if (timer == NULL) {
    timer = timerBegin(0, 80, true);
    timerAttachInterrupt(timer, &onTimer, true);
  }
  const unsigned long durationMs = 4000;
  timerAlarmWrite(timer, durationMs * 1000, false);
  timerAlarmEnable(timer);

}

void stopTimer() {
  if (timer != NULL) {
    timerAlarmDisable(timer);
    timerDetachInterrupt(timer);
    timerEnd(timer);
    timer = NULL;
  }
}

String base64_encode_utf(uint8_t * data, size_t length)
{
    size_t size = base64_encode_expected_len(length) + 1;
    char * buffer = (char *) malloc(size);
    if(buffer) {
        base64_encodestate _state;
        base64_init_encodestate(&_state);
        int len = base64_encode_block((const char *) &data[0], length, &buffer[0], &_state);
        len = base64_encode_blockend((buffer + len), &_state);
        String base64 = String(buffer);
        free(buffer);
        return base64;
    }
    return String("-FAIL-");
}
void serialEvent() {
  if (Serial.available() && !printStop) {
    printStart = true;
  }
  while (Serial.available()) {
    char inChar = (char)Serial.read();
    inputStrSerial += inChar;
    if (inChar == '\n') { 
      startTimer();  
    }
  }  
}
void sendDataWs(AsyncWebSocketClient * client, JsonDocument& doc) {
  String json = "";
  size_t len = serializeJson(doc, json);
  client->text(json);
}
void onWsEvent(AsyncWebSocket * server, AsyncWebSocketClient * client, AwsEventType type, void * arg, uint8_t *data, size_t len){
  if(type == WS_EVT_CONNECT){
    mainClient = client;
  } else if(type == WS_EVT_DISCONNECT){
    mainClient = NULL;
  } else if(type == WS_EVT_ERROR){
  } else if(type == WS_EVT_PONG){
  } else if(type == WS_EVT_DATA){  }
};  

void onPageNotFound(AsyncWebServerRequest *request) {
  request->send(404, "text/plain", "Not found");
}

void setup() {
  Serial.begin(38400);
  SPIFFS.begin();
  WiFi.softAP(ssid, password);
  ws.onEvent(onWsEvent);
  server.serveStatic("/", SPIFFS, "/").setDefaultFile("index.html");
  server.onNotFound(onPageNotFound);
  server.addHandler(&ws);
  server.begin();
}
 
void loop() {
  serialEvent();
  if(mainClient != NULL && mainClient->status() == WS_CONNECTED) {
    
    if (printStart) {
      StaticJsonDocument<32> doc;
      doc["action"] = "printStart";
      sendDataWs(mainClient, doc);
      printStop = true;
      printStart = false;
    }
    if (serialComplete) {
      StaticJsonDocument<32> doc;
      doc["action"] = "serialComplete";
      sendDataWs(mainClient, doc);
      doc["action"] = "sendImage";
      sendDataWs(mainClient, doc);      
      String answer = base64_encode_utf((uint8_t*)&inputStrSerial[0], inputStrSerial.length());
      DynamicJsonDocument docDyn(answer.length()+32);
      docDyn["image"] = answer;
      sendDataWs(mainClient, docDyn);
      doc["action"] = "completeSend";
      sendDataWs(mainClient, doc);
      printStop = false;
      printStart = false;
      serialComplete = false;
      inputStrSerial = "";
    }
  }
  ws.cleanupClients();
}
